// fileopen.cpp
// tab_size: 4
#pragma hdrfile
#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include <string.h>
#include <dos.h>

#include "fileopen.h"

/* Name of the File name property list item */
char filename_prop[] = "FILENAME";
extern HANDLE hInst;
int FAR PASCAL DialogBoxParam(HANDLE, LPSTR, HWND, FARPROC, LONG);

/* IsWild() -- Checks if the string (referenced by a NEAR pointer)
 * contains a DOS wildcard character ("*" or "?").
 * Input:
 *	psz	- String to analyze for a wildcard character.
 * Returns:
 *	TRUE if the string contains a wildcard character, FALSE otherwise.
 */
BOOL NEAR PASCAL IsWild(register PSTR psz) {
	for(;;) {
		switch (*psz++) {
			case '*':
			case '?':
				/* Found wildcard */
				return (TRUE);
			case 0:
				/* Reached end of string */
				return (FALSE);
			default:
				continue;
		}
	}
}

/* SelectFile() -- Reads the string in the edit control of the File/Open
 * dialog. If it contains a wildcard, then it attempts to fill the
 * listboxes in the File/Open dialog. Othewise it ends the dialog.
 * Modifies the FILENAME item in the property list of the window.
 * Input:
 * Returns:
 */
VOID NEAR PASCAL SelectFile(register HWND hwnd) {
	register PSTR pch;
	PSTR pch2;
	int cch;

	/* Get handle (near address) to filename data in window's
	 * property list
	 */
	pch = (PSTR) GetProp(hwnd, filename_prop);

	/* Get the text from the dialog's edit control into this address */
	GetDlgItemText(hwnd, IDD_FILENAME, pch, 64);

	if (IsWild(pch)) {
		/* Select the directory and make a listing of
		 * the directories
		 */
		DlgDirList(hwnd, (LPSTR)pch, IDD_DIRS, IDD_PATH, ATTR_DIRS);

		/* Obtain the filename-only part of the path in the edit 
		 * control
		 */
		for (pch2 = pch; *pch; pch++) {
			if (*pch == '\\' || *pch == '//' || *pch == ':') {
				pch2 = pch + 1;
			}
		}

		/* List the files in this directory based on the wildcard. */
		DlgDirList(hwnd, (LPSTR)pch2, IDD_FILES, IDD_PATH, 
								ATTR_FILES);

		/* Set the dialog's edit control to the filename part of path
		 * string.
		 */
		SetDlgItemText (hwnd, IDD_FILENAME, pch2);
	} else {
		/* The filename in the property list is not a wildcard.
		 * Remove it from the list and end dialog.
		 */
		RemoveProp(hwnd, filename_prop);
		EndDialog(hwnd, IDOK);
	}
}

/* FileOpenDlgProc() -- Dialog function for the File/Open dialog. Takes 
 * care of calling the appropriate functions for extracting the	filename 
 * and wildcard, filling the listboxes and changing the FILENAME item 
 * in the property list for the window.
 * Input:
 * Returns:
 */
BOOL FAR PASCAL FileOpenDlgProc(register HWND hwnd, WORD message,
					register WORD wParam, LONG lParam) {
	PSTR pch;

	switch (message) {
		case WM_INITDIALOG:
			/* Set the default file extension on edit window, 
			 * and try to get a listing of the files and 
			 * directories.
			 */
			SetDlgItemText(hwnd, IDD_FILENAME, (LPSTR) "*.GIF");
			SetProp(hwnd, filename_prop, LOWORD(lParam));
			SendDlgItemMessage (hwnd, IDD_FILENAME, EM_LIMITTEXT, 64, 0L);
			SelectFile(hwnd);
			break;
		case WM_COMMAND:
			switch (wParam) {
				case IDOK:
					SelectFile(hwnd);
					break;
				case IDCANCEL:
					/* Set the filename in the prop. list to NULL and quit */
					pch  = (PSTR) GetProp (hwnd, filename_prop);
					*pch = 0;
					EndDialog(hwnd, IDCANCEL);
					break;
				case IDD_FILENAME:
					/* Enable the OK button if the edit control has text. */
					EnableWindow(GetDlgItem(hwnd, IDOK), 
						GetWindowTextLength((HWND) LOWORD(lParam)));
					break;
				case IDD_FILES:
					/* The files listbox. If file selection has changed, fill
					 * the new filename into the property list buffer and set
					 * text in edit control.
					 */
					if (HIWORD(lParam) == LBN_SELCHANGE) {
						pch = (PSTR) GetProp(hwnd, filename_prop);
						DlgDirSelect(hwnd, (LPSTR) pch, IDD_FILES);
						SetDlgItemText(hwnd, IDD_FILENAME, (LPSTR) pch);
					} else if (HIWORD(lParam) == LBN_DBLCLK) {
						/* if the item was double-clicked, try to open it */
						SelectFile(hwnd);
					}
					break;
				case IDD_DIRS:
					/* The directories listbox.  Append current filename in 
					 * edit control (stripped of the path prefix) to the 
					 * name from the property list and set the new string 
					 * in the edit control.
					 */
					if (HIWORD(lParam) == LBN_SELCHANGE) {
						PSTR pch2, pchT, pchS;

						pchS = pch = (PSTR)GetProp(hwnd, filename_prop);

						/* Get the new drive/dir */
						DlgDirSelect(hwnd, pch, IDD_DIRS);
						pch2 = pch + lstrlen(pch);

						/* Fetch current contents of dialog's edit control 
						 * and append it to name from property list...
						 */
						GetDlgItemText(hwnd, IDD_FILENAME, (LPSTR) pch2, 64);

						/* ... and remove any old path information in it */
						for (pchT = pch = pch2; *pch; pch++) {
							if (*pch == '\\' || *pch == '//' || *pch == ':') {
								pchT = pch2;
							} else {
								*pchT++ = *pch;
							}
						}
						*pchT = 0;

						// Set the edit control with new string.
						SetDlgItemText(hwnd, IDD_FILENAME, (LPSTR) pchS);
					} else if (HIWORD(lParam) == LBN_DBLCLK) {
						SelectFile(hwnd);
					}
					break;
				default:
					return (FALSE);
			}
			break;
		default:
			return (FALSE);
	}
	return (TRUE);
}

// GetFilename(pstr) -- Gets a filename from the user by calling the 
// File/Open dialog.
// Input:
//	pstr	- String where we will return the name of the file selected.
//	parent	- HANDLE of the parent window to this dialog box.
// Returns:
//	Nothing.
WORD NEAR PASCAL GetFileName(PSTR pstr, HWND parent) {
	FARPROC lpfn;
	WORD temp;

	lpfn = MakeProcInstance((FARPROC) FileOpenDlgProc, hInst);
	temp = DialogBoxParam(hInst, "OpenBox", parent, lpfn, MAKELONG(pstr, 0));
	FreeProcInstance(lpfn);
	
	/* Return the return value from the dialog box. */
	return (temp);
}
