// main.cpp
// tab_size: 4
#pragma hdrfile
#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>

#include "winsupp.h"

#include <local/primer.h>

#include "ids.h"
#include "main.h"
#include "about.h"
#include "fileopen.h"
#include "ordered.h"
#include "error.h"

#define DATABUF_SIZE	10 * 1024

FILE *gif_file;
static char *databuf;

BOOL gcm_present, interlaced;
color global_cmap[256], local_cmap[256];
unsigned int lineno, image_num_colors;
screen_desc screen_descriptor;
image_desc image_descriptor;

HANDLE hInst, hDibInfo, hBitmap = NULL, hPal;
LPBITMAPINFO pDibInfo;
HWND hErrorWnd, hOutputWnd;
char szCmdLine[256];
BOOL decode_to_screen = FALSE, minimize_on_decoding = FALSE;
BOOL bFileListed = FALSE, bYieldWhileDecoding = FALSE;
BOOL gif_displayed = FALSE, bProcessing = FALSE;
dithering_type dithering = none;
char szAppName[] = "WGIF";

int get_byte() {
	return(fgetc(gif_file));
}

int out_line(unsigned char *pixels, int linelen) {
	unsigned int i, j;
	unsigned char dest_pixel;
	unsigned long line_offset;
	HDC hDC, hMemoryDC;
	HANDLE hOldBitmap;
	MSG msg;
	static unsigned int interlaced_pass;

	// Ordered dither this line before we put it out.
	switch (dithering) {
		case ordered:
			j = lineno % 16;

			for (i = 0; i < linelen; i++) {
				if (local_cmap[pixels[i]].red >= dither[j][i % 16]) {
					dest_pixel = 4;
				} else {
					dest_pixel = 0;
				}

				if (local_cmap[pixels[i]].green >= dither[j][i % 16]) {
					dest_pixel += 2;
				}

				if (local_cmap[pixels[i]].blue >= dither[j][i % 16]) {
					dest_pixel += 1;
				}
				pixels[i] = dest_pixel;
			}
			break;
		case none:
		default:
			break;
	}
	
	hDC = GetDC(hOutputWnd);
	hMemoryDC = CreateCompatibleDC(hDC);
	hOldBitmap = SelectObject(hMemoryDC, hBitmap);

	SelectPalette(hDC, hPal, 0);
	RealizePalette(hDC);

	SelectPalette(hMemoryDC, hPal, 0);
	RealizePalette(hMemoryDC);

	SetDIBits(hMemoryDC, hBitmap, image_descriptor.height - lineno - 1, 1, 
		(LPSTR) pixels, pDibInfo,  DIB_PAL_COLORS);
	
	SelectObject(hMemoryDC, hOldBitmap);
	DeleteDC(hMemoryDC);

	if (decode_to_screen && !IsIconic(hOutputWnd)) {
		SetDIBitsToDevice(hDC, 0, lineno, linelen, 1, 0, 0, 0, 1,
			(LPSTR) pixels, pDibInfo, DIB_PAL_COLORS);
	}

	ReleaseDC(hOutputWnd, hDC);

	// If we don't do this, then no other program can do anything (and
	// therefore the user can't do anything) while we are decoding.  Since
	// decoding can be such a long process, yielding to other tasks is
	// necessary.
	if (bYieldWhileDecoding) {
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Calculate where the next line of the image will be placed when
	// decoding interlaced pictures. 
	if (interlaced) {
		if (lineno == 0) {
			/* Reset the interlaced variable. */
			interlaced_pass = 1;
		}

		switch (interlaced_pass) {
			case 1:
				lineno += 8;
				if (lineno > (image_descriptor.height - 1)) {
					lineno = 4;
					interlaced_pass = 2;
				}
				break;
			case 2:
				lineno += 8;
				if (lineno > (image_descriptor.height - 1)) {
					lineno = 2;
					interlaced_pass = 3;
				}
				break;
			case 3:
				lineno += 4;
				if (lineno > (image_descriptor.height - 1)) {
					lineno = 1;
					interlaced_pass = 4;
				}
				break;
			case 4:
				lineno += 2;
				break;
		}
	} else {
		lineno++;
	}
	
	return (0);
}

// read_screen_desc() -- Reads in a screen descriptor a piece at a time
// rather than pulling in a whole structure with a fread.
// Input:
//	screen_desc	- Pointer to a screen descriptor structure to fill in.
//	gfile	- Stream to get the data from.
// Returns:
//	Always returns TRUE.  Each getw() and getc() should be checked for errors
// but I don't do it now.
//
// Note:
//	On PC's you can get away with just reading in and writing out structures
// but on many machines structures have additional padding bytes in them to
// make access to data elements faster (due to word alignment).
BOOL read_screen_desc(screen_desc *screen_descriptor, FILE *gfile) {
	screen_descriptor->width = getw(gfile);
	screen_descriptor->height = getw(gfile);
	screen_descriptor->flags = getc(gfile);
	screen_descriptor->background = getc(gfile);
	screen_descriptor->zero_terminator = getc(gfile);
	return (TRUE);
}

BOOL read_image_desc(image_desc *image_descriptor, FILE *gfile) {
	image_descriptor->image_separator = getc(gfile);
	image_descriptor->left = getw(gfile);
	image_descriptor->top = getw(gfile);
	image_descriptor->width = getw(gfile);
	image_descriptor->height = getw(gfile);
	image_descriptor->flags = getc(gfile);
	return (TRUE);
}

BOOL decode_gif(char *gif_filename) {
	char gif_signature[6], terminator_check, error_msg[256];
	unsigned int color_res, bpp, num_colors, image_bpp;
	unsigned int i;
	int nMaxWidth, nMaxHeight;
	BOOL use_gcm, verbose = FALSE;
	HDC hDC;
	WORD FAR *pDibPalette;
	LOGPALETTE *pLogPal;
	RECT WndRect, CurrentWndRect;
	extern short decoder(short linelen);

	if ((gif_file = fopen(gif_filename, "rb")) == NULL) {
		msg_error(hErrorWnd, "Unable to open file.");
		return (FALSE);
	}

	// Set up a buffer to be used for file reading.
	setvbuf(gif_file, databuf, _IOFBF, DATABUF_SIZE);
	
	// Read & verify the GIF signature on the file.
	if (fread(gif_signature, 1, 6, gif_file) != 6) {
		strcpy(error_msg, gif_filename);
		strcat(error_msg, " : Unable to read from file.");
		msg_error(hErrorWnd, error_msg);
		fclose(gif_file);
		return (FALSE);
	}

	if (strncmp(gif_signature, "GIF87a", 6) != 0) {
		strcpy(error_msg, gif_filename);
		strcat(error_msg, " : Is not a GIF file.");
		msg_error(hErrorWnd, error_msg);
		fclose(gif_file);
		return (FALSE);
	}

	if (read_screen_desc(&screen_descriptor, gif_file) == FALSE) {
		fclose(gif_file);
		return (FALSE);
	}

	// Check the flag bit in the file to see whether a global color map
	// is present.
	gcm_present = (screen_descriptor.flags & 0x80) ? TRUE : FALSE;
	color_res = ((screen_descriptor.flags & 0x70) >> 4) + 1;
	bpp = (screen_descriptor.flags & 0x07) + 1;
	num_colors = 2 << (bpp - 1);
	
	if (verbose) {
#if 0
		printf("%s:\n", gif_filename);
		printf("Screen Width = %d\nScreen Height = %d\n", 
			screen_descriptor.width, screen_descriptor.height);
	
		if (gcm_present) {
			printf("\tGlobal Color Map present\n");
		}

		printf("\tColor resolution = %d\n", color_res);
		printf("\t# Bits/pixel = %d\n", bpp);
		printf("\t# Colors = %d\n", num_colors);
	
		printf("Background color = %d\n", 
						screen_descriptor.background);
#endif

		if (screen_descriptor.zero_terminator != 0) {
			msg_error(hErrorWnd, "Error: Non zero terminator.");
			fclose(gif_file);
			return (FALSE);
		}
	}
	
	/* Read and verify the global color map if one is present. */
	if (gcm_present) {
		if (fread(global_cmap, sizeof(color), num_colors, gif_file) != 
			num_colors) {
			msg_error(hErrorWnd, "Unable to read Global Colormap.");
			fclose(gif_file);
			return (FALSE);
		}
	} else {
		/* Generate a default color map. */
		
	}

	// Some people think that it is legal to have a GIF file without 
	// an image.  According to the chart on page four of the original 
	// GIF spec though, the sequence Image Descriptor-Local Color Map
	// -Raster Data is repeated ONE to N times.
	if ((terminator_check = fgetc(gif_file)) == ';') {
		msg_error(hErrorWnd, "Encountered GIF terminator without image!");
		fclose(gif_file);
		return (FALSE);
	}

	do {
		// We haven't encountered a terminator yet so we need to push
		// the character we read to check back onto the stream.
		if (ungetc(terminator_check, gif_file) == EOF) {
			msg_error(hErrorWnd, "Unable to push back terminator "
				"check char.");
			fclose(gif_file);
			return (TRUE);
		}
		
		// We need to skip any extension blocks that may appear at 
		// this point.
	
		if (read_image_desc(&image_descriptor, gif_file) == FALSE) {
			fclose(gif_file);
			return (FALSE);
		}

		// At this point we should really check to see that the size 
		// of this picture is in fact <= the screen dimensions 
		// specified in the screen descriptor.  According to the GIF 
		// spec, if it is larger then the file is in error.

		use_gcm = (image_descriptor.flags & 0x80) ? FALSE : TRUE;
		interlaced = (image_descriptor.flags & 0x40) ? TRUE : FALSE;
	
		if (!use_gcm) {
			image_bpp = image_descriptor.flags & 0x07;
		} else {
			image_bpp = bpp;
		}
		image_num_colors = 2 << (image_bpp - 1);

		if (verbose) {
#if 0
			printf("Image Left = %d\nImage Top = %d\n", 
				image_descriptor.left, image_descriptor.top);
			printf("Image Width = %d\nImage Height = %d\n", 
				image_descriptor.width, image_descriptor.height);
			printf(use_gcm ? "\tUse Global Color Map\n" : 
				"\tUse Local Color Map\n");
			printf(interlaced ? "\tImage is Interlaced\n" : 
				"\tImage is sequential\n");

			if (!use_gcm) {
				printf("\t# Bits/pixel (local) = %d\n", 
					image_bpp);
			}

			if (image_descriptor.image_separator == ',') {
				printf("Found Image separator\n");
			} else {
				msg_error(hErrorWnd, "Image separator expected but "
					"not found.");
				fclose(gif_file);
				return (FALSE);
			}
#endif
		}
			
		if (!use_gcm) {
			// We need to read in the Local Color Map.
			if (fread(&local_cmap, sizeof(color), image_num_colors, 
				gif_file) != image_num_colors) {
				msg_error(hErrorWnd, "Read size smaller than expected "
					"reading local color map.");
				fclose(gif_file);
				return (FALSE);
			}
		} else {
			// We want to always use the local color palette.  So we
			// should copy the global to local at this point.
			for (i = 0; i < num_colors; i++) {
				local_cmap[i] = global_cmap[i];
			}
		}

		// Allocate the DIBInfo needed as a destination for the image.
		if ((hDibInfo = GlobalAlloc(GMEM_FIXED,	sizeof(BITMAPINFOHEADER) + 
			(256 * sizeof(RGBQUAD)))) == NULL) {
			msg_error(hErrorWnd, "Unable to allocate DIB header.");
			fclose(gif_file);
			return (FALSE);
		}
		pDibInfo = (LPBITMAPINFO) GlobalLock(hDibInfo);
		pDibInfo->bmiHeader.biSize = (DWORD) sizeof(BITMAPINFOHEADER);
		pDibInfo->bmiHeader.biWidth = image_descriptor.width;
		pDibInfo->bmiHeader.biHeight = image_descriptor.height;
		pDibInfo->bmiHeader.biPlanes = 1;
		pDibInfo->bmiHeader.biBitCount = 8;
		pDibInfo->bmiHeader.biCompression = BI_RGB;
		pDibInfo->bmiHeader.biSizeImage = 0L;
		pDibInfo->bmiHeader.biXPelsPerMeter = 0L;
		pDibInfo->bmiHeader.biYPelsPerMeter = 0L;
		pDibInfo->bmiHeader.biClrUsed = 256;
		pDibInfo->bmiHeader.biClrImportant = 0;

		pDibPalette = (WORD FAR *) &(pDibInfo->bmiColors[0]);
		for (i = 0; i < image_num_colors; i++) {
			pDibPalette[i] = (WORD) i;
		}

		hDC = GetDC(hOutputWnd);

		if ((pLogPal = (LOGPALETTE *) LocalAlloc(LMEM_FIXED, 
			sizeof(LOGPALETTE) + (image_num_colors * 
			sizeof(PALETTEENTRY)))) == NULL) {
			// Failed to allocate enough memory for the logical palette.
			msg_error(hErrorWnd, "Out of memory: "
				"Allocating logical palette");
			fclose(gif_file);
			return (FALSE);
		}

		pLogPal->palVersion = 0x300;
		switch (dithering) {
			case ordered:
				pLogPal->palNumEntries = 8;

				// Set up a logical palette for ordered dithering.
				for (i = 0; i < 8; i++) {
					pLogPal->palPalEntry[i].peRed = (i & 4) ? 255 : 0;
					pLogPal->palPalEntry[i].peGreen = (i & 2) ? 255 : 0;
					pLogPal->palPalEntry[i].peBlue = (i & 1) ? 255 : 0;
					pLogPal->palPalEntry[i].peFlags = (BYTE) 0;
				}
				break;
			case none:
			default:
				pLogPal->palNumEntries = image_num_colors;

				for (i = 0; i < image_num_colors; i++) {
					pLogPal->palPalEntry[i].peRed = local_cmap[i].red;
					pLogPal->palPalEntry[i].peGreen = local_cmap[i].green;
					pLogPal->palPalEntry[i].peBlue = local_cmap[i].blue;
					pLogPal->palPalEntry[i].peFlags = (BYTE) 0;
				}
				break;
		}
		
		hPal = CreatePalette(pLogPal);
		LocalFree((HANDLE) pLogPal);

		hBitmap = CreateDIBitmap(hDC, (LPBITMAPINFOHEADER) 
			&(pDibInfo->bmiHeader), 0, NULL, NULL, 0);
		ReleaseDC(hOutputWnd, hDC);

		// Users don't like to keep having to Zoom the window.  If the user
		// has it zoomed, leave the window size alone.
		if (!IsZoomed(hOutputWnd)) {
			// Figure out what the new size of the window should be, taking
			// into acccount things like menus and title bars, etc.
			GetWindowRect(hOutputWnd, (LPRECT) &CurrentWndRect);
			WndRect.left = WndRect.top = 0;
			WndRect.bottom = WndRect.top + image_descriptor.height;
			WndRect.right = WndRect.left + image_descriptor.width;
			AdjustWindowRect((LPRECT) &WndRect, WS_OVERLAPPEDWINDOW, TRUE);

			// Find out the size of the visible screen and make sure that our
			// window is not sized in such a way that it extends beyond the
			// edges.
			nMaxWidth = GetDeviceCaps(hDC, HORZRES) - CurrentWndRect.left;
			nMaxHeight = GetDeviceCaps(hDC, VERTRES) - CurrentWndRect.top;
				
			// Resize the window to accomodate the new image size.
			MoveWindow(hOutputWnd, CurrentWndRect.left, CurrentWndRect.top, 
				min(abs(WndRect.right - WndRect.left), nMaxWidth),
				min(abs(WndRect.bottom - WndRect.top), nMaxHeight), TRUE);

			// Remove the scroll bars from the window.
			SetScrollRange(hOutputWnd, SB_HORZ, 0, 0, FALSE);
			SetScrollRange(hOutputWnd, SB_VERT, 0, 0, FALSE);
		}

		if (minimize_on_decoding) {
			SendMessage(hOutputWnd, WM_SYSCOMMAND, SC_MINIMIZE, 0L);
		}

		lineno = 0;

		if (decoder(image_descriptor.width) != 0) {
			// Got an error back from the decoder.
			msg_error(hErrorWnd, "Unable to decode this GIF.");
			fclose(gif_file);
			return (FALSE);
		}

		// The following is to fix an unfortunate bug in the GIF
		// decoder.  It normally leaves the terminating zero on the
		// last Raster Data Block sitting on the stream.  Thus we have
		// to pick it up by hand.
		fgetc(gif_file);
	} while ((terminator_check = fgetc(gif_file)) != ';');

	fclose(gif_file);
	return (TRUE);
}

void display_gif(HWND hWnd) {
	HDC hDC, hMemoryDC;
	HANDLE hOldBitmap;
	BITMAP bitmap;

	hDC = GetDC(hWnd);

	hMemoryDC = CreateCompatibleDC(hDC);
	hOldBitmap = SelectObject(hMemoryDC, hBitmap);
	SetMapMode(hMemoryDC, GetMapMode(hDC));
	
	SelectPalette(hDC, hPal, 0);
	RealizePalette(hDC);

	SelectPalette(hMemoryDC, hPal, 0);
	RealizePalette(hMemoryDC);

	GetObject(hBitmap, sizeof(BITMAP), (LPSTR) &bitmap);

	BitBlt(hDC, 0, 0, bitmap.bmWidth, bitmap.bmHeight, hMemoryDC, 0, 0, 
		SRCCOPY);

	DeleteDC(hMemoryDC);

	ReleaseDC(hWnd, hDC);
}

// shutdown() -- Cleans up after the program and sees to saving user options.
// Input:
//	gif_displayed	- Flag to signal that we need to do some memory cleanup.
// Returns:
//	Nothing.
void shutdown(BOOL gif_displayed) {
	char temp[10];

	// Delete vagrant memory structures.
	if (gif_displayed) {
		DeleteObject(hPal);
		DeleteObject(hBitmap);
		GlobalFree(hDibInfo);
	}

	// Save various user option settings.
	WriteProfileString(szAppName, "decode_to_screen", itoa((int) 
		decode_to_screen, temp, 10));
	WriteProfileInt(szAppName, "dithering", (int) dithering);
	WriteProfileInt(szAppName, "minimize_on_decoding", 
		(int) minimize_on_decoding);
	WriteProfileInt(szAppName, "bYieldWhileDecoding", 
		(int) bYieldWhileDecoding);

	PostQuitMessage(0);
}

// LoadGIF() -- 
void LoadGIF(HWND hWnd, char *szGIFFile) {
	char szNewTitle[40];

	// bProcessing turns on/off the ability to close this
	// program using the CLOSE item on the system menu.
	bProcessing = TRUE;

	if (gif_displayed) {
		DeleteObject(hPal);
		DeleteObject(hBitmap);
		GlobalFree(hDibInfo);

		/* Clean the screen of the old image. */
		gif_displayed = FALSE;
		InvalidateRect(hWnd, (LPRECT) NULL, TRUE);
		SetWindowText(hWnd, szAppName);
	}

	hOutputWnd = hWnd;
	ChangeUMall(hWnd, "BusyMenu");
	if (decode_gif(szGIFFile)) {
		gif_displayed = TRUE;

		InvalidateRect(hWnd, (LPRECT) NULL, FALSE);
							
		// Change the window title to include the name of
		// the picture.  It is an extra clue to the user
		// about the state of the program and it tells the
		// user who has the program minimized that we are
		// finished decoding.
		strcpy(szNewTitle, szAppName);
		strcat(szNewTitle, " - ");
		strcat(szNewTitle, szGIFFile);
		SetWindowText(hWnd, szNewTitle);
	}
	ChangeUMall(hWnd, NULL);

	// Re-gray the menu item.
	GrayUMon(hWnd, IDM_RELOAD);
	bProcessing = FALSE;
}

// gif_wnd() -- Processes messages for the main GIF window.
// Input:
// Returns:
LONG FAR PASCAL gif_wnd(HWND hWnd, unsigned message, WORD wParam,
	LONG lParam) {
	int num_pal_entries_changed;
	static char gif_filename[80], temp_filename[80], szMessage[80];
	FARPROC lpProcAbout, lpProcInfo;
	HDC hDC, hMemoryDC;
	PAINTSTRUCT ps;
	HANDLE hOldBitmap, hOldPal, hBusyMenu, hIcon;
	static WORD wWidth = 0, wHeight = 0;
	DWORD dwStyle;
	static int nVScrollPos, nHScrollPos;
	RECT ClientRect;
	BOOL bBug = FALSE;

	switch (message) {
		case WM_CREATE:
			decode_to_screen = (BOOL) GetProfileInt(szAppName, 
				"decode_to_screen", (int) TRUE);
			CheckMenuItem(GetMenu(hWnd), IDM_DECODE_TO_SCREEN, 
				decode_to_screen ? MF_CHECKED : MF_UNCHECKED);
			
			dithering = (dithering_type) GetProfileInt(szAppName, 
				"dithering", (int) FALSE);
			switch (dithering) {
				case ordered:
					CheckMenuItem(GetMenu(hWnd), IDM_ORDERED_DITHERING, 
						MF_CHECKED);
					break;
				case none:
				default:
					CheckMenuItem(GetMenu(hWnd), IDM_NONE, MF_CHECKED);
					break;
			}

			minimize_on_decoding = (BOOL) GetProfileInt(szAppName,
				"minimize_on_decoding", (int) FALSE);
			CheckMenuItem(GetMenu(hWnd), IDM_MINIMIZE_ON_DECODING,
				minimize_on_decoding ? MF_CHECKED : MF_UNCHECKED);

			bYieldWhileDecoding = (BOOL) GetProfileInt(szAppName,
				"bYieldWhileDecoding", (int) FALSE);
			CheckMenuItem(GetMenu(hWnd), IDM_YIELD_WHILE_DECODING,
				bYieldWhileDecoding ? MF_CHECKED : MF_UNCHECKED);

			// Remove the scroll bars from the window.
			SetScrollRange(hWnd, SB_HORZ, 0, 0, FALSE);
			SetScrollRange(hWnd, SB_VERT, 0, 0, FALSE);

			if (bFileListed) {
				// The user added an argument to the line used to call WGIF.
				// We are going to assume that this argument is the name of
				// a file to load and attempt to do just that.
				strcpy(gif_filename, szCmdLine);
			}
			break;
		case WM_COMMAND:
			// A command from the application menu.
			switch (wParam) {
				case IDM_OPEN:
					if (GetFileName((PSTR) temp_filename, hWnd) == IDOK) {
						strcpy(gif_filename, temp_filename);
						
						LoadGIF(hWnd, gif_filename);
						SendMessage(hWnd, WM_SIZE, SIZENORMAL, 0L);
					}
					break;
				case IDM_RELOAD:
					LoadGIF(hWnd, gif_filename);
					break;
				case IDM_EXIT:
					shutdown(gif_displayed);
					break;
				case IDM_NONE:
					if (gif_displayed && (dithering != none)) {
						// The user has changed dithering options so turn
						// on the automatic reload feature.
						GrayUMoff(hWnd, IDM_RELOAD);
					}

					if (dithering == ordered) {
						dithering = none;
						CheckMenuItem(GetMenu(hWnd), IDM_NONE, MF_CHECKED);
						CheckMenuItem(GetMenu(hWnd), IDM_ORDERED_DITHERING, 
							MF_UNCHECKED);
					}
					break;
				case IDM_ORDERED_DITHERING:
					if (gif_displayed) {
						// The user has changed dithering options so turn
						// on the automatic reload feature.
						GrayUMoff(hWnd, IDM_RELOAD);
					}

					// Toggle the menu item and the internal flag.
					if (dithering == ordered) {
						dithering = none;
						CheckMenuItem(GetMenu(hWnd), IDM_NONE, MF_CHECKED);
						CheckMenuItem(GetMenu(hWnd), IDM_ORDERED_DITHERING, 
							MF_UNCHECKED);
					} else {
						dithering = ordered;
						CheckMenuItem(GetMenu(hWnd), IDM_NONE, MF_UNCHECKED);
						CheckMenuItem(GetMenu(hWnd), IDM_ORDERED_DITHERING,
							MF_CHECKED);
					}
					break;
				case IDM_DECODE_TO_SCREEN:
					// Toggle the menu item and the internal flag.
					if (decode_to_screen) {
						decode_to_screen = FALSE;
						CheckMenuItem(GetMenu(hWnd), IDM_DECODE_TO_SCREEN, 
							MF_UNCHECKED);
					} else {
						decode_to_screen = TRUE;
						CheckMenuItem(GetMenu(hWnd), IDM_DECODE_TO_SCREEN,
							MF_CHECKED);
					}
					break;
				case IDM_MINIMIZE_ON_DECODING:
					if (minimize_on_decoding) {
						minimize_on_decoding = FALSE;
						CheckMenuItem(GetMenu(hWnd), IDM_MINIMIZE_ON_DECODING,
							MF_UNCHECKED);
					} else {
						minimize_on_decoding = TRUE;
						CheckMenuItem(GetMenu(hWnd), IDM_MINIMIZE_ON_DECODING,
							MF_CHECKED);
					}
					break;
				case IDM_YIELD_WHILE_DECODING:
					if (bYieldWhileDecoding) {
						bYieldWhileDecoding = FALSE;
						CheckMenuItem(GetMenu(hWnd), IDM_YIELD_WHILE_DECODING,
							MF_UNCHECKED);
					} else {
						bYieldWhileDecoding = TRUE;
						CheckMenuItem(GetMenu(hWnd), IDM_YIELD_WHILE_DECODING,
							MF_CHECKED);
					}
					break;
				case IDM_ABOUT:
					lpProcAbout = MakeProcInstance((FARPROC) about_dlg, 
						hInst);
					DialogBox(hInst, "ABOUTDLG", hWnd, lpProcAbout);
					FreeProcInstance(lpProcAbout);
					break;
				default:
					/* Let Windows process it. */
					return (DefWindowProc(hWnd, message, wParam, lParam));
			}
			break;
		case WM_VSCROLL:
			switch (wParam) {
				case SB_LINEUP:
					nVScrollPos--;
					break;
				case SB_LINEDOWN:
					nVScrollPos++;
					break;
				case SB_PAGEUP:
					nVScrollPos -= wHeight / 4;
					break;
				case SB_PAGEDOWN:
					nVScrollPos += wHeight / 4;
					break;
				case SB_THUMBPOSITION:
					nVScrollPos = LOWORD(lParam);
					break;
				default:
					break;
			}
			nVScrollPos = max(0, min(nVScrollPos, (int) wHeight));

			if (nVScrollPos != GetScrollPos(hWnd, SB_VERT)) {
				SetScrollPos(hWnd, SB_VERT, nVScrollPos, TRUE);
				InvalidateRect(hWnd, (LPRECT) NULL, FALSE);
			}
			break;
		case WM_HSCROLL:
			switch (wParam) {
				case SB_LINEUP:
					nHScrollPos--;
					break;
				case SB_LINEDOWN:
					nHScrollPos++;
					break;
				case SB_PAGEUP:
					nHScrollPos -= wWidth / 4;
					break;
				case SB_PAGEDOWN:
					nHScrollPos += wWidth / 4;
					break;
				case SB_THUMBPOSITION:
					nHScrollPos = LOWORD(lParam);
					break;
				default:
					break;
			}
			nHScrollPos = max(0, min(nHScrollPos, (int) wWidth));

			if (nHScrollPos != GetScrollPos(hWnd, SB_HORZ)) {
				SetScrollPos(hWnd, SB_HORZ, nHScrollPos, TRUE);
				InvalidateRect(hWnd, (LPRECT) NULL, FALSE);
			}
			break;
		case WM_PAINT:
			if (gif_displayed) {
				hDC = BeginPaint(hWnd, &ps);
				hMemoryDC = CreateCompatibleDC(hDC);
				hOldBitmap = SelectObject(hMemoryDC, hBitmap);
				SetMapMode(hMemoryDC, GetMapMode(hDC));

				// The same palette needs to be selected into both the source
				// and destination DC's for the upcoming blit.
				SelectPalette(hDC, hPal, 0);
				RealizePalette(hDC);

				SelectPalette(hMemoryDC, hPal, 0);
				RealizePalette(hMemoryDC);

				BitBlt(hDC, ps.rcPaint.left, ps.rcPaint.top, 
					ps.rcPaint.right - ps.rcPaint.left, 
					ps.rcPaint.bottom - ps.rcPaint.top, 
					hMemoryDC, ps.rcPaint.left + GetScrollPos(hWnd, SB_HORZ), 
					ps.rcPaint.top + GetScrollPos(hWnd, SB_VERT), SRCCOPY);

				DeleteDC(hMemoryDC);
				EndPaint(hWnd, &ps);
			} else {
				return (DefWindowProc(hWnd, message, wParam, lParam));
			}
			break;
		case WM_QUERYNEWPALETTE:
			if (gif_displayed) {
				// We are about to become the active window.  We need
				// to set our palette appropriately and redraw the
				// window.
				hDC = GetDC(hWnd);
				hOldPal = SelectPalette(hDC, hPal, 0);
				num_pal_entries_changed = RealizePalette(hDC);
				ReleaseDC(hWnd, hDC);

				if (num_pal_entries_changed > 0) {
					// Invalidate the entire client area.
					InvalidateRect(hWnd, (LPRECT) NULL, FALSE);
				}
			}
			break;
		case WM_PALETTECHANGED:
			// Check to make sure we didn't cause the message!  If we didn't
			// then update the client area to try to make the image look the
			// best it can.
			if ((wParam != hWnd) && gif_displayed) {
				hDC = GetDC(hWnd);
				hOldPal = SelectPalette(hDC, hPal, 0);
				num_pal_entries_changed = RealizePalette(hDC);
				
				if (num_pal_entries_changed > 0) {
					UpdateColors(hDC);
				}
				ReleaseDC(hWnd, hDC);
			}
			break;
		case WM_SIZE:
			// The user has resized the window so we need to check and see if
			// scroll bars should be added or deleted from the window.
			if (((wParam == SIZENORMAL) || (wParam == SIZEFULLSCREEN))
				&& gif_displayed && !IsIconic(hWnd)) {
				// Get the size of the client rectangle area.
				dwStyle = GetWindowLong(hWnd, GWL_STYLE);
				GetClientRect(hWnd, &ClientRect);

				// Compare the size of the window to the size of the image
				// that needs to appear in it.
				if ((ClientRect.right + 1) < image_descriptor.width) {
					wWidth = image_descriptor.width - ClientRect.right;
				} else {
					wWidth = 0;
				}

				if ((ClientRect.bottom + 1) < image_descriptor.height) {
					wHeight = image_descriptor.height - ClientRect.bottom;
				} else {
					wHeight = 0;
				}

				SetScrollRange(hWnd, SB_HORZ, 0, wWidth, TRUE);
				SetScrollRange(hWnd, SB_VERT, 0, wHeight, TRUE);

				InvalidateRect(hWnd, (LPRECT) NULL, FALSE);
			} else {
				// Pass it on if unproccessed.
				return (DefWindowProc(hWnd, message, wParam, lParam));
			}
			break;
		case WM_CLOSE:
			if (!bProcessing) {
				DestroyWindow(hWnd);
			}
			break;
		case WM_DESTROY:
			shutdown(gif_displayed);
			break;
		default:
			// Pass it on if unproccessed.
			return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	return (NULL);
}

// InitApplication() -- Initializes window data and registers window class.
// Input:
//	hInstance	- Handle to this instance of the application.
// Returns:
//	FALSE if we could not register the new window class, TRUE otherwise.
//
// Note:
//	This function is called at initialization time only if no other 
// instances of the application are running.  This function performs 
// initialization tasks that can be done once for any number of running 
// instances.  
//
//	In this case, we initialize a window class by filling out a data 
// structure of type WNDCLASS and calling the Windows RegisterClass() 
// function.  Since all instances of this application use the same 
// window class, we only need to do this when the first instance is 
// initialized.
BOOL InitApplication(HANDLE hInstance) {
	WNDCLASS wc;

	// Fill in window class structure with parameters that describe the
	// main window.
	wc.style = NULL;
	wc.lpfnWndProc = gif_wnd;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance, (LPSTR) "WGIF_ICON");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(BLACK_BRUSH); 
	wc.lpszMenuName =  "GIFMainMenu";
	wc.lpszClassName = "GIFWinClass";

	// Register the window class and return success/failure code.
	return ((RegisterClass(&wc) ? TRUE : FALSE));
}

// InitInstance() -- Saves instance handle and creates main window.
// Input:
//	hInstance	-
//	nCmdShow	- 
// Returns:
//	TRUE if the main window was created, FALSE otherwise.
//
// Note:
//	This function is called at initialization time for every instance of 
// this application.  This function performs initialization tasks that 
// cannot be shared by multiple instances.  
//
//	In this case, we save the instance handle in a static variable and 
// create and display the main program window.
BOOL InitInstance(HANDLE hInstance, int nCmdShow) {
	HWND hWnd;

	// Save the instance handle in static variable, which will be used in
	// many subsequence calls from this application to Windows.
	hInst = hInstance;
	if ((databuf = (char *) malloc(DATABUF_SIZE)) == NULL) {
		return (FALSE);
	}

	// Create a main window for this application instance.
	hWnd = CreateWindow("GIFWinClass", szAppName, 
		WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL, 
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
		NULL, NULL, hInstance, NULL);

	// If window could not be created, return "failure".
	if (!hWnd) {
		return (FALSE);
	}

	hErrorWnd = hWnd;

	// Make the window visible; update its client area; and return "success".
	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	// If a file was named on the command line, load it now.
	if (bFileListed) {
		SendMessage(hWnd, WM_COMMAND, IDM_RELOAD, 0L);
	}
	return (TRUE);
}

// WinMain() -- Calls initialization functions, processes message loop.
// Input:
//	hInstance	- 
//	hPrevInstance	- 
//	lpCmdLine	- 
//	nCmdShow	- 
// Returns:
//	NULL if there is an error during initialization, the value returned in
// the wParam of the last GetMessage call otherwise.
//	
// Note:
//	Windows recognizes this function by name as the initial entry point 
// for the program.  This function calls the application initialization 
// routine, if no other instance of the program is running, and always 
// calls the instance initialization routine.  It then executes a message 
// retrieval and dispatch loop that is the top-level control structure 
// for the remainder of execution.  The loop is terminated when a WM_QUIT 
// message is received, at which time this function exits the application 
// instance by returning the value passed by PostQuitMessage(). 
//
//	If this function must abort before entering the message loop, it 
// returns the conventional value NULL.  
int PASCAL _export WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
	LPSTR lpCmdLine, 
	int nCmdShow) {
	MSG msg;

	// Look for an already running instance of this program.
	if (!hPrevInstance) {
		// One does not exist so we need to register window classes
		// and perform all manner of initialization that will be
		// needed by all instances of this application.
		if (!InitApplication(hInstance)) {
			return (NULL);
		}
	}

	// Copy the command line so that we can potentially act upon any files
	// that may be listed there.
	if (lpCmdLine[0] != 0) {
		// There are additional arguments on the command line.
		bFileListed = TRUE;
		lstrcpy(szCmdLine, lpCmdLine);
	}

	// Perform initializations that apply to this instance only.
	if (!InitInstance(hInstance, nCmdShow)) {
		return (NULL);
	}

	// Main message dispatch loop.
	while (GetMessage(&msg, NULL, NULL, NULL)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return (msg.wParam);
}
